@@ -3,7 +3,7 @@
 # Univention AD Connector
 #  Basic class for the UCS connector part
 #
-# SPDX-FileCopyrightText: 2004-2025 Univention GmbH
+# SPDX-FileCopyrightText: 2004-2026 Univention GmbH
 # SPDX-License-Identifier: AGPL-3.0-only
 
 
@@ -18,6 +18,7 @@ import re
 import sqlite3 as lite
 import string
 import sys
+import traceback
 from logging import getLogger
 from types import FunctionType
 
@@ -251,6 +252,38 @@ class configdb:
                     self._dbcon.close()
                 self._dbcon = lite.connect(self.filename)
 
+    def get_rows_with_key_ending_in_dn(self, section, dn):
+        ending = '%%,%s' % dn
+        for _i in [1, 2]:
+            try:
+                cur = self._dbcon.cursor()
+                cur.execute("SELECT * FROM '%s' WHERE key LIKE ?" % section, (ending,))  # noqa: S608
+                rows = cur.fetchall()
+                cur.close()
+                return rows or []
+            except lite.Error:
+                # TODO: why is this just a warning?
+                log.warning('sqlite error', exc_info=True)
+                if self._dbcon:
+                    self._dbcon.close()
+                self._dbcon = lite.connect(self.filename)
+
+    def get_rows_with_value_ending_in_dn(self, section, dn):
+        ending = '%%,%s' % dn
+        for _i in [1, 2]:
+            try:
+                cur = self._dbcon.cursor()
+                cur.execute("SELECT * FROM '%s' WHERE value LIKE ?" % section, (ending,))  # noqa: S608
+                rows = cur.fetchall()
+                cur.close()
+                return rows or []
+            except lite.Error:
+                # TODO: why is this just a warning?
+                log.warning('sqlite error', exc_info=True)
+                if self._dbcon:
+                    self._dbcon.close()
+                self._dbcon = lite.connect(self.filename)
+
 
 class RFC4514_dn:
     special_dn_chars = '"+,;<=>'
@@ -565,6 +598,9 @@ class ucs:
     def _remove_rejected_ucs(self, filename):
         self._remove_config_option('UCS rejected', filename)
 
+    def _save_ad_reject_reason(self, id, msg):
+        self._set_config_option('AD rejected reason', str(id), msg)
+
     def list_rejected_ucs(self, filter_noresync=False):
         rejected = self._get_config_items('UCS rejected')
         if filter_noresync:
@@ -640,6 +676,22 @@ class ucs:
         dn = self._get_dn_by_con(dn_con)
         return self.dn_mapped_to_base(dn, self.lo.base)
 
+    def _update_subtree_dns_in_mappings_by_con(self, old_con_dn, old_ucs_dn, new_con_dn, new_ucs_dn):
+        old_con_dn_lower = old_con_dn.lower()
+        old_ucs_dn_lower = old_ucs_dn.lower()
+        new_con_dn_lower = new_con_dn.lower()
+        new_ucs_dn_lower = new_ucs_dn.lower()
+        for row in self.config.get_rows_with_key_ending_in_dn('DN Mapping CON', old_con_dn_lower):
+            _new_key = self._subtree_replace(row[0], old_con_dn_lower, new_con_dn_lower)
+            _new_val = self._subtree_replace(row[1], old_ucs_dn_lower, new_ucs_dn_lower)
+            self._remove_config_option('DN Mapping CON', row[0])
+            self._set_config_option('DN Mapping CON', _new_key, _new_val)
+        for row in self.config.get_rows_with_value_ending_in_dn('DN Mapping UCS', old_con_dn_lower):
+            _new_key = self._subtree_replace(row[0], old_ucs_dn_lower, new_ucs_dn_lower)
+            _new_val = self._subtree_replace(row[1], old_con_dn_lower, new_con_dn_lower)
+            self._remove_config_option('DN Mapping UCS', row[0])
+            self._set_config_option('DN Mapping UCS', _new_key, _new_val)
+
     def _check_dn_mapping(self, dn_ucs, dn_con):
         dn_con_mapped = self._get_dn_by_ucs(dn_ucs.lower())
         dn_ucs_mapped = self._get_dn_by_con(dn_con.lower())
@@ -653,6 +705,13 @@ class ucs:
         prefix = '[%14s] [%10s] %r' % (property_type or '?', obj.get('modtype', '?'), obj.get('dn', '?'))
         return '{}: {}{}'.format(direction, prefix, f': {message}' if message else '')
 
+    def check_syncmode_ucs(self, property_key):
+        # if sync is read (sync from AD) or none, there is nothing to do
+        if self.property[property_key].sync_mode in ('read', 'none'):
+            log.info('sync_from_ucs ignored, sync_mode is %s', self.property[property_key].sync_mode)
+            return False
+        return True
+
     def __sync_file_from_ucs(self, filename, append_error=''):
         """sync changes from UCS stored in given file"""
         try:
@@ -690,6 +749,10 @@ class ucs:
         _attr = new or old
         _mod, key = self.identify_udm_object(dn, _attr)
 
+        if not key or not self.check_syncmode_ucs(key):
+            log.info('No mapping was found for dn: %s', dn)
+            return True
+
         if not new:
             change_type = "delete"
             log.info("__sync_file_from_ucs: object was deleted")
@@ -825,7 +888,7 @@ class ucs:
 
     def get_ucs_ldap_object_dn(self, dn):
         try:
-            return self.lo.lo.lo.search_s(dn, ldap.SCOPE_BASE, '(objectClass=*)', ('dn',))[0][0]
+            return self.lo.lo.lo.search_s(dn, ldap.SCOPE_BASE, '(objectClass=*)', ('1.1',))[0][0]
         except ldap.NO_SUCH_OBJECT:
             return
         except ldap.INVALID_DN_SYNTAX:
@@ -1093,7 +1156,7 @@ class ucs:
 
         MAPPING = self.property[property_type]
         for attr_key, attributes in MAPPING.attributes.items():
-            if attributes.sync_mode not in ['read', 'sync']:
+            if attributes.sync_mode not in ('read', 'sync'):
                 log.debug("__set_values: Skip %s mode attribute %s ", attributes.sync_mode, attr_key)
                 continue
 
@@ -1119,7 +1182,7 @@ class ucs:
             return
         for attr_key, post_attributes in MAPPING.post_attributes.items():
             log.trace("__set_values: mapping for attribute: %s", attr_key)
-            if post_attributes.sync_mode not in ['read', 'sync']:
+            if post_attributes.sync_mode not in ('read', 'sync'):
                 log.trace("__set_values: Skip %s mode attribute %s ", post_attributes.sync_mode, attr_key)
                 continue
 
@@ -1183,25 +1246,19 @@ class ucs:
         serverctrls = [PostReadControl(True, ['entryUUID', 'entryCSN'])]
         res = ucs_object.modify(serverctrls=serverctrls, response=response)
         if res:
-            object['dn'] = res
             for c in response.get('ctrls', []):
                 if c.controlType == PostReadControl.controlType:  # If the modify actually did something
                     entryUUID = c.entry['entryUUID'][0].decode('ASCII')
                     entryCSN = c.entry['entryCSN'][0].decode('ASCII')
                     self._remember_entryCSN_commited_by_connector(entryUUID, entryCSN)
-            res = True
         return res
 
     def move_in_ucs(self, property_type, object, module, position):
-        if self.lo.compare_dn(object['olddn'].lower(), object['dn'].lower()):
-            log.warning("move_in_ucs: cancel move, old and new dn are the same (%r to %r)", object['olddn'], object['dn'])
-            return True
-
         log.debug("move_in_ucs: move object from %r to %r", object['olddn'], object['dn'])
         ucs_object = univention.admin.objects.get(module, None, self.lo, dn=object['olddn'], position='')
         ucs_object.open()
-        ucs_object.move(object['dn'])
-        return True
+        res = ucs_object.move(object['dn'])
+        return res
 
     def _get_entryUUID(self, dn):
         try:
@@ -1294,70 +1351,22 @@ class ucs:
         :param property_type:
             the type of the object to be synced, must be part of the mapping. (e.g. "user", "group", "dc", "windowscomputer", etc.)
         :param object:
-            A dictionary describing the AD object.
+            A dictionary describing the AD object as returned by _object_mapping()
             modtype: A modification type ("add", "modify", "move", "delete")
             dn: The DN of the object in the UCS-LDAP
             olddn: The olddn of the object object in UCS-LDAP (e.g. on "move" operation)
         :ptype object: dict
         :param pre_mapped_ad_dn:
-            pass
+            same as original_object['dn']
         :param original_object:
-            pass
+            as returned by __object_from_element()
         """
         # NOTE: pre_mapped_ad_dn means: original ad_dn (i.e. before _object_mapping)
         # this function gets an object from the ad class, which should be converted into a ucs module
-
-        # if sync is write (sync to AD) or none, there is nothing to do
-        if not property_type or self.property[property_type].sync_mode in ['write', 'none']:
-            if property_type:
-                log.debug(self.context_log(property_type, object, "sync ignored: sync_mode is %s", to_ucs=True), self.property[property_type].sync_mode)
-            else:
-                log.debug(self.context_log(property_type, object, "sync ignored: no mapping defined", to_ucs=True))
+        if not self.check_syncmode_ad(property_type, object):
             return True
-
-        if object['dn'].find('\\0ACNF:') > 0:
-            log.process('Ignore conflicted object: %s', object['dn'])
-            return True
-
         try:
-            guid = decode_guid(original_object.get('attributes').get('objectGUID')[0])
-
-            object['changed_attributes'] = []
-            if object['modtype'] == 'modify' and original_object:
-                old_ad_object = self.adcache.get_entry(guid)
-                log.debug("sync_to_ucs: old_ad_object: %s", old_ad_object)
-                log.debug("sync_to_ucs: new_ad_object: %s", original_object['attributes'])
-                original_attributes = original_object['attributes']
-                if old_ad_object:
-                    for attr in original_object['attributes']:
-                        if old_ad_object.get(attr) != original_attributes.get(attr):
-                            object['changed_attributes'].append(attr)
-                    for attr in old_ad_object:
-                        if old_ad_object.get(attr) != original_attributes.get(attr) and attr not in object['changed_attributes']:
-                            object['changed_attributes'].append(attr)
-                    if not (set(object['changed_attributes']) - self.irrelevant_attributes):
-                        if (
-                            property_type == "user"
-                            and self.configRegistry.is_false('connector/ad/mapping/user/password/disabled', True)
-                            and not self.configRegistry.is_true('connector/ad/mapping/user/password/kinit', False)
-                        ):
-                            if object['attributes'].get('pwdLastSet', [b'1'])[0] == b'0':
-                                log.debug("sync_to_ucs: pwdLastSet is 0. Do not ignore %r", original_object['dn'])
-                            else:
-                                log.debug("sync_to_ucs: ignore %r", original_object['dn'])
-                                return True
-                        else:
-                            log.debug("sync_to_ucs: ignore %r", original_object['dn'])
-                            log.trace("sync_to_ucs: changed_attributes=%s", object['changed_attributes'])
-                            return True
-                else:
-                    object['changed_attributes'] = list(original_attributes.keys())
-            log.debug("The following attributes have been changed: %s", object['changed_attributes'])
-
-            result = False
-
-            # Check if the object on UCS side should be synchronized
-            #  https://forge.univention.org/bugzilla/show_bug.cgi?id=37351
+            guid = decode_guid(object.get('attributes').get('objectGUID')[0])
             if property_type == "windowscomputer":
                 old_ucs_ldap_object = {}
                 old_ucs_ldap_object['dn'] = object.get('olddn', object['dn'])
@@ -1369,12 +1378,14 @@ class ucs:
 
             old_object = self.get_ucs_object(property_type, object.get('olddn', object['dn']))
 
+            # Corrections in case the target situation is unexpected:
             if old_object and object['modtype'] == 'add':
                 object['modtype'] = 'modify'
-            if not old_object and object['modtype'] == 'modify':
-                object['modtype'] = 'add'
-            if not old_object and object['modtype'] == 'move':
-                object['modtype'] = 'add'
+            if not old_object and object['modtype'] in ['modify', 'move']:
+                if self.get_ucs_ldap_object_dn(object['dn']):
+                    object['modtype'] = 'modify'
+                else:
+                    object['modtype'] = 'add'
 
             if self.group_member_mapping_cache_ucs.get(object['dn'].lower()) and object['modtype'] != 'delete':
                 self.group_member_mapping_cache_ucs[object['dn'].lower()] = None
@@ -1404,17 +1415,70 @@ class ucs:
                 self._remove_dn_mapping(object['dn'], pre_mapped_ad_dn)
                 self.adcache.remove_entry(guid)
             if object['modtype'] == 'move':
-                result = self.move_in_ucs(property_type, object, module, position)
-                self._remove_dn_mapping(object['olddn'], '')  # we don't know the old ad-dn here anymore, will be checked by remove_dn_mapping
-                self._check_dn_mapping(object['dn'], pre_mapped_ad_dn)
+                if self.lo.compare_dn(object['olddn'].lower(), object['dn'].lower()):
+                    log.warning("sync_to_ucs: cancel move, old and new dn are the same (%r to %r)", object['olddn'], object['dn'])
+                    result = True
+                else:
+                    result = self.move_in_ucs(property_type, object, module, position)
+                    old_con_dn = original_object.get('olddn')
+                    if old_con_dn and pre_mapped_ad_dn != old_con_dn:
+                        self._remove_dn_mapping(object['olddn'], old_con_dn)
+                        self._update_group_related_caches(
+                            property_type,
+                            old_con_dn=old_con_dn,
+                            old_ucs_dn=object['olddn'],
+                            new_con_dn=pre_mapped_ad_dn,
+                            new_ucs_dn=object['dn'],
+                        )
+                    # self._remove_dn_mapping(object['olddn'], '')  # we don't know the old ad-dn here anymore, will be checked by remove_dn_mapping
+                    self._check_dn_mapping(object['dn'], pre_mapped_ad_dn)
 
             if object['modtype'] == 'modify':
-                result = self.modify_in_ucs(property_type, object, module, position)
+
+                modified_dn = self.modify_in_ucs(property_type, object, module, position)
+                modified_parent_dn = self.lo.parentDn(modified_dn).lower()
+                parent_object_dn = self.lo.parentDn(object['dn']).lower()
+                moved_or_renamed = False
+
+                if not self.lo.compare_dn(modified_parent_dn, parent_object_dn):
+                    # additionally move the object if position changed
+                    log.info('move object from %r to %r', modified_dn, object['dn'])
+                    ucs_object = univention.admin.objects.get(module, None, self.lo, dn=modified_dn, position='')
+                    ucs_object.open()
+                    ucs_object.move(object['dn'])
+                    moved_or_renamed = True
+                elif object['dn'].lower() != modified_dn.lower():
+                    # e.g. username was changed in UCS but the AD dn remained the same
+                    object['dn'] = modified_dn
+                    moved_or_renamed = True
+                elif 'olddn' in object:
+                    # simple rename (e.g. ou's)
+                    if object['olddn'].lower() != object['dn'].lower():
+                        moved_or_renamed = True
+
+                if moved_or_renamed:
+                    # update group cache
+                    old_con_dn = original_object.get('olddn', pre_mapped_ad_dn)
+                    old_dn = object.get('olddn', object['dn'])
+                    self._update_group_related_caches(
+                        property_type,
+                        old_con_dn=old_con_dn,
+                        old_ucs_dn=old_dn,
+                        new_con_dn=pre_mapped_ad_dn,
+                        new_ucs_dn=object['dn'],
+                    )
+
+                    if not self.lo.compare_dn(old_dn, object['dn']):
+                        self._remove_dn_mapping(old_dn, old_con_dn)
+
+                # Finally commit the current DNs to the DN mapping cache
                 self._check_dn_mapping(object['dn'], pre_mapped_ad_dn)
                 self.adcache.add_entry(guid, original_object.get('attributes'))
+                result = True
 
             if not result:
                 log.warning("Failed to get Result for DN (%r)", object['dn'])
+                self._save_ad_reject_reason(object['attributes']['uSNChanged'][0].decode(), "Failed to get Result")
                 return False
 
             if object['modtype'] in ['add', 'modify']:
@@ -1428,14 +1492,17 @@ class ucs:
 
         except univention.admin.uexceptions.valueInvalidSyntax as msg:
             log.error("InvalidSyntax: %s (%r)", msg, object['dn'])
+            self._save_ad_reject_reason(object['attributes']['uSNChanged'][0].decode(), f"InvalidSyntax\n{msg}")
             return False
         except univention.admin.uexceptions.valueMayNotChange as msg:
             log.error("Value may not change: %s (%r)", msg, object['dn'])
+            self._save_ad_reject_reason(object['attributes']['uSNChanged'][0].decode(), f"Value may not change\n{msg}")
             return False
         except ldap.SERVER_DOWN:
             raise
-        except Exception:  # FIXME: which exception is to be caught?
+        except Exception as msg:  # FIXME: which exception is to be caught?
             log.exception("Unknown Exception during sync_to_ucs")
+            self._save_ad_reject_reason(object['attributes']['uSNChanged'][0].decode(), f"{msg}\n{traceback.format_exc()}")
             return False
 
     @staticmethod
@@ -1444,14 +1511,14 @@ class ucs:
         return ldap.dn.str2dn(dn.lower())[-len(x):] == x
 
     @staticmethod
-    def _subtree_replace(dn, subtree, subtreereplace):
+    def _subtree_replace(dn, subtree, subtreereplace, case_folding=True):
         extra = ''
         if subtree.startswith(',') and subtreereplace.startswith(','):
             subtreereplace = subtreereplace[1:]
             subtree = subtree[1:]
             extra = ','
-        _dn = ldap.dn.str2dn(dn.lower())
-        _subtree = ldap.dn.str2dn(subtree.lower())
+        _dn = ldap.dn.str2dn(dn.lower() if case_folding else dn)
+        _subtree = ldap.dn.str2dn(subtree.lower() if case_folding else subtree)
         if _dn[-len(_subtree):] != _subtree or (extra and _dn == _subtree):
             return dn
         return ldap.dn.dn2str(ldap.dn.str2dn(dn)[:-len(_subtree)] + ldap.dn.str2dn(subtreereplace))
